/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2002 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* This code is mostly adapted from CoolPlayer's CPI_Player_CoDec_MPEG.c:
 * CoolPlayer - Blazing fast audio player.
 *   Copyright (C) 2000-2001 Niek Albers
 *   Copyright (C) 2001 Robert Leslie
 *
 *   Also released under the terms of the GPL v.2
 *
 * Because Robert Leslie pointed me to it to get an reference implementaion
 *   - THANKS!.
 *
 * Feel free to use it as mid-layer C++ API!
 *
 */

/* Short Description:
 *   Midlevel and C++ API for the libMAD !!! (mp3 library)
 */

#ifndef MAD_HELPER_H
#define MAD_HELPER_H

#include "config.hh"

#if __MP3__ == 1

#include <iostream>
#include <stdio.h>
#include <string>
#include <deque>

#include <unistd.h>

// mutex
#include <sigc++/thread.h>

// I WANT A REAL OO only OS !!!!!!!!!!!

extern "C" {
#include "mad.h" // GPL mp3 lib
}


/***********************
 * A dither class hack *
 ***********************/

class Dither
{
public:
  
  inline unsigned long Prng () {
    return (prng_state * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
  }

  inline signed int DoDither (mad_fixed_t sample)
  {
    mad_fixed_t output, mask, random;
    const unsigned int scalebits = MAD_F_FRACBITS + 1 - 24; // we use 24 bit here

#ifdef DO_DITHER
    /* noise shape */
    sample += dither_error[0] - dither_error[1] + dither_error[2];
    
    dither_error[2] = dither_error[1];
    dither_error[1] = dither_error[0] / 2;
    
    /* bias */
    output = sample + (1L << (scalebits - 1) );
    mask = (1L << (scalebits + 1) ) - 1;
    
    /* dither */
    random  = Prng ();
    output += (random & mask) - (dither_random & mask);
    
    //dither_random = random;
    
    /* clip */
    const mad_fixed_t t_max = MAD_F_ONE - 1;
    const mad_fixed_t t_min = -MAD_F_ONE;
    
    if (output > t_max) {
      output = t_max;
      
      if (sample > t_max)
	sample = t_max;
    }
    else if (output < t_min) {
      output = t_min;
      
      if (sample < t_min)
	sample = t_min;
    }
    
    /* quantize */
    output &= ~mask;
    
    /* error feedback */
    dither_error[0] = sample - output;
    
    /* scale */
    return output >> scalebits;
#endif
    return sample >> scalebits;
  }
  
protected:
  unsigned long prng_state;
  mad_fixed_t dither_error[3];
  mad_fixed_t dither_random;
};

class MADHelper
{
public:
  
  // frame start table (needed for efficent seek) / can be shared
  // between MADHelpers ...
  struct FrameCache {
    
    FrameCache () {
      std::cout << "FrameCache: CTor: " << this << std::endl;
      used = 0;
    }
    
    ~FrameCache () {
      std::cout << "FrameCache: DTor: " << this << std::endl;
      if (used > 0)
	std::cout << "~FrameCache: BUG: in dtor and used > 0 ... !?!" << std::endl;
    }
    
    void Use () {
      SigC::Threads::MLock a (lock);
      ++ used;
    }
    
    void Unuse () {
      SigC::Threads::MLock a (lock);
      -- used;
    }
    
    SigC::Threads::Mutex lock;
    unsigned int used;
    
    std::deque <long> frames;
  };
  
  MADHelper ();
  ~MADHelper ();
  
  bool Open (const std::string& fname, MADHelper* cache_partner = 0);
  bool Close ();
  
  long Length ();
  int Tracks ();
  int Rate ();
  
  bool Read (int length, mad_fixed_t* buffer);
  bool Seek (long pos);
  
protected:
  
  void Cleanup ();
  bool ParseXing (struct mad_bitptr ptr, unsigned int bitlen);
  bool ScanHeader ();
  
  bool DecodeNext (bool frame_too);
  
  // scale MAD output
  void PackPCM (mad_fixed_t* pcm, unsigned int nsamples, mad_fixed_t const* ch1, mad_fixed_t const* ch2);
  
  std::string m_fname;
  int m_fd;                        /* stream handle */
  
  long m_size;                     /* stream size, in bytes */
  
  signed long m_frames;            /* number of current decoded frame */
  long m_byte_of_mp3frame;         /* byte pos in mp3-stream */
  int m_sample_of_mp3frame;        /* pcm_frame of current mp3-frame */
  int m_pcm_frames;                /* total pcm_frames in file */
  
  int m_samplerate;                /* samplerate */
  int m_tracks;                    /* tracks / channels */
  
  struct xing {
    long flags;
    unsigned long frames;
    unsigned long bytes;
    unsigned char toc [100];
    long scale;
  };
  xing m_xing;                     /* Xing VBR tag data */
  
  mad_stream m_stream;             /* MAD stream structure */
  mad_frame m_frame;               /* MAD frame structure */
  mad_synth m_synth;               /* MAD synth structure */
  
  unsigned int m_samplecount;      /* samples output from current frame */
  
  mad_timer_t m_timer;             /* current playing time position */
  mad_timer_t m_length;            /* total playing time of current stream */
  unsigned long m_rate;            /* bitrate sum for computing average */
  
  FrameCache* m_frame_cache;
  
  unsigned int m_buflen;           /* input stream buffer vaild length */
  unsigned char m_buffer [40000];  /* input stream buffer - Rene: main src */ 
  
  Dither left_dither, right_dither;
};

#ifdef TO_BE_USED

enum {
  XING_FRAMES = 0x00000001L,
  XING_BYTES  = 0x00000002L,
  XING_TOC    = 0x00000004L,
  XING_SCALE  = 0x00000008L
};

#endif // TO_BE_USED

#endif // __MP3__

#endif // MAD_HELPER_H
